#include "crf.h"
#include <cmath>

using namespace std;

double SGD_ETA0 = 0.1;
double SGD_ITER = 30;
double SGD_ALPHA = 0.85;
double SGD_TAU = 1.0;

static double
l1norm(const vector<double>& v)
{
  double sum = 0;
  for (int i = 0; i < v.size(); i++) sum += abs(v[i]);
  return sum;
}

inline void apply_l1_penalty(const int i, const double u,
			     vector<double> & _vl,
			     vector<double> & q)
{
  double & w = _vl[i];
  const double z = w;
  double & qi = q[i];
  if (w > 0) { 
    w = max(0.0, w - (u + qi));
  } else if (w < 0) {
    w = min(0.0, w + (u - qi));
  }
  qi += w - z;
}

int
CRF_Model::perform_SGD()
{
  cerr << "performing SGD" << endl;

  const double l1param = _inequality_width;
  const int d = _fb.Size();

  vector<int> ri(_vs.size());
  for (int i = 0; i < ri.size(); i++) ri[i] = i;

  vector<int> edge_ids;
  for (int j = 0; j < _num_label_clusters; j++)
    for (int k = 0; k < _num_label_clusters; k++)
      edge_ids.push_back(edge_feature_id(j, k));

  vector<double> grad(d);
  int iter = 0, iter_sample = 0, m;
  const double eta0 = SGD_ETA0;
  const double tau = SGD_TAU * _vs.size();

  cerr << "l1param = " << l1param << endl;
  cerr << "eta0 = " << eta0 << endl;

  double u = 0;
  vector<double> q(d, 0);

  for (int iter = 0; iter < SGD_ITER; iter++) {

    random_shuffle(ri.begin(), ri.end());

    int n = 0, ncorrect = 0, ntotal = 0;
    double logl = 0;
    for (int i0 = 0; i0 < _vs.size(); i0++, iter_sample++) {

      const Sequence & seq = _vs[ri[i0]];
      ntotal += seq.vs.size();

      // compute model / emprical expectation
      initialize_edge_weights();
      const double fp = forward_backward(seq);
      const double clogl = calc_loglikelihood(seq);
      logl += clogl;

      double eta = eta0 / (1.0  + iter_sample / tau);
      if (SGD_ALPHA > 0) {
	eta = eta0 * pow(SGD_ALPHA, (double)iter_sample / _vs.size()); // exponential decay
      }

      u += eta * l1param;

      for (size_t i = 0; i < seq.vs.size(); i++) {
	const Sample & s = seq.vs[i];

	// states
	const vector<double> wsum = calc_state_weight(i);
	for (vector<int>::const_iterator j = s.positive_features.begin(); j != s.positive_features.end(); j++){
	  for (vector<int>::const_iterator k = _feature2mef[*j].begin(); k != _feature2mef[*j].end(); k++) {
	    const double me = wsum[_fb.Feature(*k).label()];
	    const double ee = (_fb.Feature(*k).label() == seq.vs[i].label) ? 1 : 0;
	    const double grad = me - ee;
	    _vl[*k] -= eta * grad;
	    apply_l1_penalty(*k ,u, _vl, q);
	  }
	}
	if (s.label == max_element(wsum.begin(), wsum.end()) - wsum.begin()) ncorrect++;
	if (i == seq.vs.size() - 1) continue;

	// edges
	const int c0 = label2cluster[seq.vs[i].label];
	const int c1 = label2cluster[seq.vs[i+1].label];
	for (int j = 0; j < _num_label_clusters; j++) {
	  const double lhs = forward_cache_clustered(i, j);
	  for (int k = 0; k < _num_label_clusters; k++) {
	    const double rhs = backward_cache_clustered(i+1, k);
	    assert(lhs != DBL_MAX && rhs != DBL_MAX);
	    const int id = edge_feature_id(j, k);
	    const double me = lhs * edge_weight(j, k) * rhs;
	    const double ee = ( (j == c0) && (k == c1) ) ? 1 : 0;
	    const double grad = me - ee;
	    _vl[id] -= eta * grad;
	  }
	}
      }

      // for edges
      if (l1param > 0) {
	for (vector<int>::const_iterator i = edge_ids.begin(); i != edge_ids.end(); i++) {
	  apply_l1_penalty(*i, u, _vl, q);
	}
      }
      
    }
    logl /= _vs.size();
    
    fprintf(stderr, "%4d logl = %8.3f acc = %6.4f ", iter, logl, (double)ncorrect / ntotal);


    double heldout_logl = 0;
    if (_nheldout > 0) {
      heldout_logl = heldout_likelihood();
      fprintf(stderr, "  heldout_logl = %8.3f, acc = %6.4f\n", heldout_logl, 1 - _heldout_error);
    } else {
      cerr << endl;
    }
    if (l1param > 0) {
      const double l1 = l1norm(_vl); // this is not accurate if updates are lazy
      const double f = logl - l1param * l1;
      int nonzero = 0;
      for (size_t j = 0; j < d; j++) if (_vl[j] != 0) nonzero++;
    }
  }


  return 0;
}

